# Fuzz Tests

This directory contains all the Fuzz tests for cudf library.


## Steps to write a fuzz test

1. Add a Data Handler class which actually generates the necessary random data according to your requirements. This class should be added in `cudf/cudf/testing/`. A sample data handler class is: `CSVWriter`: https://github.com/rapidsai/cudf/blob/branch-0.16/python/cudf/cudf/testing/csv.py
2. Data Handlers are registered by the `pythonfuzz` decorator. At runtime, the Fuzzer will continuously run registered fuzz tests.
  
```python
from cudf.testing.csv import CSVWriter

@pythonfuzz(data_handle=CSVWriter)
def csv_writer_test(data_from_generate_input):
    ...
    ...
    ...

if __name__ == "__main__":
    ...
    ...

```
## Steps to run fuzz tests

1. To run a fuzz test, for example a test(method) is in `write_csv.py`:

```bash
python write_csv.py your_function_name
```

To run a basic csv write test in `write_csv.py`:
```bash
python write_csv.py csv_writer_test
```

## Tips to run specific crash file/files

Using the `pythonfuzz` decorator pass in `regression=True` with `dirs` having list of directories 
```python
@pythonfuzz(data_handle=CSVWriter, regression=True, dir=["/cudf/python/cudf/cudf/_fuzz_testing"])
```


## Tips to run for varying parameter combinations

In the `pythonfuzz` decorator you can pass in the function parameters you would like to pass to the
fuzz-test being written via `params` as a dictionary. The values in dictionary are sampled randomly
and passed to the `your_custom_fuzz_test`.

If a parameter value depends the kind of input generated by the `data_handle`(in this case `CSVReader`),
then you can assign `ALL_POSSIBLE_VALUES` constant to it. This constant is used as an identifier by the
`data_handle` to generate random parameter values for that specific parameter purely based on data.
To perform this customization `set_rand_params` should be implemented as shown in the below example. 
```python
from cudf._fuzz_testing.main import pythonfuzz
from cudf._fuzz_testing.utils import ALL_POSSIBLE_VALUES
@pythonfuzz(
    data_handle=CSVWriter,
    params={
        "columns": ALL_POSSIBLE_VALUES,
        "is_folder": [True, False, None],
        "chunksize": ALL_POSSIBLE_VALUES,
    },
)
def your_custom_fuzz_test(data_from_data_handle, dtype, is_folder, header):
    ...
    ...
    ...
```

A sample implementation of `set_rand_params` in a `data_handle` class:
```
def set_rand_params(self, params):
    params_dict = {}
    for param, values in params.items():
        if values == ALL_POSSIBLE_VALUES:
            if param == "columns":
                col_size = self._rand(len(self._current_buffer.columns))
                params_dict[param] = list(
                    np.unique(
                        np.random.choice(
                            self._current_buffer.columns, col_size
                        )
                    )
                )
            elif param == "chunksize":
                params_dict[param] = np.random.choice(
                    [
                        None,
                        np.random.randint(
                            low=1, high=max(1, len(self._current_buffer))
                        ),
                    ]
                )
        else:
            params_dict[param] = np.random.choice(values)
    self._current_params["test_kwargs"] = self.process_kwargs(params_dict)
```